// tiny modification to calculate timer2 reload values for different MCU clock speeds - see the lines below including F_CPU, July 2022 ceptimus
// Phil_G's lockdown project, May 2020.
// Simple 4 Channel R/C Receiver with frequency hopping and programmable failsafe on all functions
// Use an RF-Nano or Promini 8mhz or 16mhz, 1S Li-Ion to Vcc and servo positives or if >5v use Vin. All negs common.
// Must add a big capacitor (say 100uF) across 0v and 3v3 otherwise the NRF24L01 can misbehave.
// Servo signals on pins D2, D3, D4, D5. 5v LED + to A0 - to A1 blinks on 20 packets, stays on if failsafe
// Set failsafe remotely or by grounding pin D8, when set with rx pin, servo wiggles to confirm its stored.
// If failsafe is not set, loss of signal reverts to defaults - low throttle and neutral controls.
// Choose a unique 'pipeIn' value and use same value in tx, I suggest the last 5 digits of your phone number
// Choose 16 random hopping channels spread out between 0 and 125
#include <avr/wdt.h> // built-in watch dog timer functions
#include <EEPROM.h>
#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>
#include <Servo.h>
#define failSafePin 8 // was 7 on early protoypes, use 8 to leave 6 & 7 free for extra servos
#define thr_failsafe 1100

// timer 2 OCR2A values (1024 prescaler) for 5ms and 6ms interrupt intervals
#define TOP_5ms (((F_CPU >> 10) + 100) / 200)
#define TOP_6ms (((F_CPU >> 10) + 83) / 167)

Servo ch1, ch2, ch3, ch4;

struct Packet {
  unsigned int aileron  = 1500;
  unsigned int elevator = 1500;
  unsigned int throttle = thr_failsafe;
  unsigned int rudder   = 1500;
};

Packet data;
unsigned long newChannelTime = 0, lastRecvTime = 0, now = 0;
volatile bool timeout = 0, hop = 1;
unsigned int int5ms = TOP_5ms, int6ms = TOP_6ms;  // These will be 78 and 94 respectively for a 16MHz crystal, but adjust automatically to suit processor clock speed

/************* Radio config - make it unique! **********************************/
const byte pipeIn[] = "12345";   // last 5 digits of your phone number is sufficiently unique in our small group
const byte mychannels[] = {38, 8, 68, 78, 43, 28, 53, 3, 23, 48, 33, 58, 63, 13, 73, 18}; // choose 16 RANDOM but non-adjacent channels 1 to 82.  Don't use these!
/*******************************************************************************/

RF24 radio(10, 9);  // RF-NANO and new builds, ce=10, csn=9.   Phils original was RF24 radio(9, 10) dont use this for new builds except LGT8F RF-NANO
int rfchan = 0, missedcounter = 0, flashcount = 0; // NRF channel index, number of packets missed

void setup()
{
  wdt_disable();
  WDTCSR |= 0b00011000;           // change enable & reset enable
  WDTCSR = 0b00001000 |  WDTO_2S; // 2 second timeout interval
  wdt_reset();

  pinMode(failSafePin, INPUT_PULLUP); // momentarily link to gnd to set failsafe
  pinMode(A0, OUTPUT); // led
  pinMode(A1, OUTPUT); // led
  digitalWrite(A0, 0); // 5v led pos
  digitalWrite(A1, 0); // led ground

  // Servo pins
  ch1.attach(2);
  ch2.attach(3);
  ch3.attach(4);
  ch4.attach(5);
  ch1.write(data.aileron); ch2.write(data.elevator); ch3.write(data.throttle); ch4.write(data.rudder);   // Servo outputs

  //Configure the NRF24 module
  radio.begin(); delayMicroseconds(500);
  radio.openReadingPipe(1, pipeIn);  delayMicroseconds(500);
  radio.setAutoAck(false); delayMicroseconds(500);
  radio.setDataRate(RF24_250KBPS); delayMicroseconds(500);
  radio.setPALevel(RF24_PA_MIN); delayMicroseconds(500);
  radio.setChannel(mychannels[0]); delayMicroseconds(500);
  radio.startListening();

  // set up 5ms interrupt
  TCCR2A = 0;        // reset Timer2 control reg A
  TCCR2A |= (1 << WGM21);     // set WGM21 bit, CTC mode
  TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20); // 1024 prescaler
  OCR2A = int6ms;                 // top = 94 for 6ms, top = 78 for 5ms with 16 MHz crystal, but adjusts to suit processor clock speed
  TIMSK2 |= (1 << OCIE2A);    // enable timer2 compare interrupt
  sei();        // enable global interrupts
}

void loop()  {
  wdt_reset();                 // reset the WDT
  TCNT2 = 0; timeout = 0;      // reset timer1 to zero
  while (timeout == 0) {
    if (radio.available() ) {  // is there a packet on the current RF channel?
      TCNT2 = 0; OCR2A = int6ms; timeout = 0;    // reset timer1 and set compare value to 6ms x16mhz/8 = 12000
      radio.read(&data, sizeof(Packet));
      radio.flush_rx();
      if (data.aileron > 2500) {
        data.aileron -= 2000;
        save_fs();
      }
      ch1.write(data.aileron); ch2.write(data.elevator); ch3.write(data.throttle); ch4.write(data.rudder);   // Servo outputs
      if (++flashcount == 10) {
        flashcount = 0;
        PORTC ^= 1;
      }
      hop = 1;
      missedcounter = 0;
      rfchan = (rfchan + hop) % 16; radio.setChannel(mychannels[rfchan]); // cycle through mychannels[]
    }
  }

  // 5 or 6ms timeout
  if (++missedcounter < 7) {  // if we've missed 1 to 5 packets, hop on 5ms at a time until we see data
    TCNT2 = 0; OCR2A = int5ms; timeout = 0;    // reset timer1 and set compare value to 5ms x16mhz/8 = 10000
    hop = 1;
    rfchan = (rfchan + hop) % 16; radio.setChannel(mychannels[rfchan]); // try next channel
  }

  else {  // we missed six packets on the trot, restart from zero
    rfchan = 0; hop = 0;
    radio.setChannel(mychannels[rfchan]);
    radio.flush_rx();
    TCNT2 = 0; OCR2A = 255; timeout = 0;    // set timer2 long and set compare value to 6ms x16mhz/1024 = 94
  }

  if (missedcounter > 40) {  // failsafe if nothing heard for 1 second
    failSafe(); // Signal lost.. revert to failsafe values, either saved or default
  }

  if (digitalRead(failSafePin) == 0) {
    save_fs();
    while (digitalRead(failSafePin) == 0) {
      wdt_reset();     // reset the WDT
      ch1.write(1400); // wiggle to say 'saved'
      ch2.write(1400); // wiggle to say 'saved'
      delay(200);
      ch1.write(1600); // wiggle to say 'saved'
      ch2.write(1600); // wiggle to say 'saved'
      delay(200);
    }
  }
  radio.flush_rx(); // clear anything received during failsafe setting
}


// save failsafe positions
void save_fs() {
    EEPROMWriteInt(0, 0x5AA5);  // flag failsafe set
    EEPROMWriteInt(2, data.aileron);
    EEPROMWriteInt(4, data.elevator);
    EEPROMWriteInt(6, data.throttle);
    EEPROMWriteInt(8, data.rudder);
}

// This function will write a 2 byte integer to the eeprom at the specified address and address + 1
void EEPROMWriteInt(int p_address, int p_value)
{
  byte lowByte = p_value % 256;
  byte highByte = p_value / 256;
  EEPROM.write(p_address, lowByte);
  EEPROM.write(p_address + 1, highByte);
}

//This function will read a 2 byte integer from the eeprom at the specified address and address + 1
unsigned int EEPROMReadInt(int p_address)
{
  byte lowByte = EEPROM.read(p_address);
  byte highByte = EEPROM.read(p_address + 1);
  return lowByte + highByte * 256;
}

void failSafe()
{
  rfchan = 0; hop = 0;
  PORTC |= 1; // led on
  if (EEPROMReadInt(0) == 0x5AA5) {   // if failsafe set
    data.aileron  = EEPROMReadInt(2); // ailerons
    data.elevator = EEPROMReadInt(4); // elevator
    data.throttle = EEPROMReadInt(6); // throttle
    data.rudder   = EEPROMReadInt(8); // rudder
  }
  else {
    data.aileron  = 1500; // Neutral
    data.elevator = 1500; // Neutral
    data.throttle = thr_failsafe; // Low throttle
    data.rudder   = 1500; // Neutral
  }
  ch1.write(data.aileron); ch2.write(data.elevator); ch3.write(data.throttle); ch4.write(data.rudder);   // Servo outputs
}

ISR(TIMER2_COMPA_vect) {   // timer, set to 6ms whilst in sync, 5ms when resyncing
  timeout = 1;
}

// If you use this software in whole or in part please acknowledge its source - Phil_G on most forums, philg@talk21.com  http://www.singlechannel.co.uk
// Read the story here: http://mode-zero.uk/viewtopic.php?f=42&t=971    and its predecessor here: http://mode-zero.uk/viewtopic.php?f=42&t=945
